@@ -4,26 +4,18 @@ import enum
 import functools
 import inspect
 import operator
+import os
+import sys
+import time
+import traceback
 import types
-from typing import (
-    Any,
-    Callable,
-    Dict,
-    Iterable,
-    List,
-    Mapping,
-    Optional,
-    Set,
-    Tuple,
-    TypeVar,
-    Union,
-    cast,
-)
+from typing import Any, Callable, Iterable, Mapping, Optional, TypeVar, Union, cast
 
 from typing_extensions import ParamSpec
 
 from snowflake import connector
-from snowflake.connector import telemetry as connector_telemetry, time_util
+from snowflake.connector import connect, telemetry as connector_telemetry, time_util
+from snowflake.ml import version as snowml_version
 from snowflake.ml._internal import env
 from snowflake.ml._internal.exceptions import (
     error_codes,
@@ -46,10 +38,42 @@ _Args = ParamSpec("_Args")
 _ReturnValue = TypeVar("_ReturnValue")
 
 
+def _get_login_token() -> Union[str, bytes]:
+    with open("/snowflake/session/token") as f:
+        return f.read()
+
+
+def _get_snowflake_connection() -> Optional[connector.SnowflakeConnection]:
+    conn = None
+    if os.getenv("SNOWFLAKE_HOST") is not None and os.getenv("SNOWFLAKE_ACCOUNT") is not None:
+        try:
+            conn = connect(
+                host=os.getenv("SNOWFLAKE_HOST"),
+                account=os.getenv("SNOWFLAKE_ACCOUNT"),
+                token=_get_login_token(),
+                authenticator="oauth",
+            )
+        except Exception:
+            # Failed to get a new SnowflakeConnection in SPCS. Fall back to using the active session.
+            # This will work in some cases once SPCS enables multiple authentication modes, and users select any auth.
+            pass
+
+    if conn is None:
+        try:
+            active_session = next(iter(session._get_active_sessions()))
+            conn = active_session._conn._conn if active_session.telemetry_enabled else None
+        except snowpark_exceptions.SnowparkSessionException:
+            # Failed to get an active session. No connection available.
+            pass
+
+    return conn
+
+
 @enum.unique
 class TelemetryProject(enum.Enum):
     MLOPS = "MLOps"
     MODELING = "ModelDevelopment"
+    EXPERIMENT_TRACKING = "ExperimentTracking"
     # TODO: Update with remaining projects.
 
 
@@ -75,6 +99,8 @@ class TelemetryField(enum.Enum):
     KEY_FUNC_PARAMS = "func_params"
     KEY_ERROR_INFO = "error_info"
     KEY_ERROR_CODE = "error_code"
+    KEY_STACK_TRACE = "stack_trace"
+    KEY_DURATION = "duration"
     KEY_VERSION = "version"
     KEY_PYTHON_VERSION = "python_version"
     KEY_OS = "operating_system"
@@ -94,13 +120,13 @@ class _TelemetrySourceType(enum.Enum):
     AUGMENT_TELEMETRY = "SNOWML_AUGMENT_TELEMETRY"
 
 
-_statement_params_context_var: contextvars.ContextVar[Dict[str, str]] = contextvars.ContextVar("statement_params")
+_statement_params_context_var: contextvars.ContextVar[dict[str, str]] = contextvars.ContextVar("statement_params")
 
 
 class _StatementParamsPatchManager:
     def __init__(self) -> None:
-        self._patch_cache: Set[server_connection.ServerConnection] = set()
-        self._context_var: contextvars.ContextVar[Dict[str, str]] = _statement_params_context_var
+        self._patch_cache: set[server_connection.ServerConnection] = set()
+        self._context_var: contextvars.ContextVar[dict[str, str]] = _statement_params_context_var
 
     def apply_patches(self) -> None:
         try:
@@ -112,7 +138,7 @@ class _StatementParamsPatchManager:
         except snowpark_exceptions.SnowparkSessionException:
             pass
 
-    def set_statement_params(self, statement_params: Dict[str, str]) -> None:
+    def set_statement_params(self, statement_params: dict[str, str]) -> None:
         # Only set value if not already set in context
         if not self._context_var.get({}):
             self._context_var.set(statement_params)
@@ -147,7 +173,6 @@ class _StatementParamsPatchManager:
                 if throw_on_patch_fail:  # primarily used for testing
                     raise
                 # TODO: Log a warning, this probably means there was a breaking change in Snowpark/SnowflakeConnection
-                pass
 
     def _patch_with_statement_params(
         self, target: object, function_name: str, param_name: str = "statement_params"
@@ -192,10 +217,10 @@ class _StatementParamsPatchManager:
 
         setattr(target, function_name, wrapper)
 
-    def __getstate__(self) -> Dict[str, Any]:
+    def __getstate__(self) -> dict[str, Any]:
         return {}
 
-    def __setstate__(self, state: Dict[str, Any]) -> None:
+    def __setstate__(self, state: dict[str, Any]) -> None:
         # unpickling does not call __init__ by default, do it manually here
         self.__init__()  # type: ignore[misc]
 
@@ -205,7 +230,7 @@ _patch_manager = _StatementParamsPatchManager()
 
 def get_statement_params(
     project: str, subproject: Optional[str] = None, class_name: Optional[str] = None
-) -> Dict[str, Any]:
+) -> dict[str, Any]:
     """
     Get telemetry statement parameters.
 
@@ -226,8 +251,8 @@ def get_statement_params(
 
 
 def add_statement_params_custom_tags(
-    statement_params: Optional[Dict[str, Any]], custom_tags: Mapping[str, Any]
-) -> Dict[str, Any]:
+    statement_params: Optional[dict[str, Any]], custom_tags: Mapping[str, Any]
+) -> dict[str, Any]:
     """
     Add custom_tags to existing statement_params.  Overwrite keys in custom_tags dict that already exist.
     If existing statement_params are not provided, do nothing as the information cannot be effectively tracked.
@@ -241,7 +266,7 @@ def add_statement_params_custom_tags(
     """
     if not statement_params:
         return {}
-    existing_custom_tags: Dict[str, Any] = statement_params.pop(TelemetryField.KEY_CUSTOM_TAGS.value, {})
+    existing_custom_tags: dict[str, Any] = statement_params.pop(TelemetryField.KEY_CUSTOM_TAGS.value, {})
     existing_custom_tags.update(custom_tags)
     # NOTE: This can be done with | operator after upgrade from py3.8
     return {
@@ -284,17 +309,17 @@ def get_function_usage_statement_params(
     *,
     function_category: str = TelemetryField.FUNC_CAT_USAGE.value,
     function_name: Optional[str] = None,
-    function_parameters: Optional[Dict[str, Any]] = None,
+    function_parameters: Optional[dict[str, Any]] = None,
     api_calls: Optional[
-        List[
+        list[
             Union[
-                Dict[str, Union[Callable[..., Any], str]],
+                dict[str, Union[Callable[..., Any], str]],
                 Union[Callable[..., Any], str],
             ]
         ]
     ] = None,
-    custom_tags: Optional[Dict[str, Union[bool, int, str, float]]] = None,
-) -> Dict[str, Any]:
+    custom_tags: Optional[dict[str, Union[bool, int, str, float]]] = None,
+) -> dict[str, Any]:
     """
     Get function usage statement parameters.
 
@@ -316,12 +341,12 @@ def get_function_usage_statement_params(
         >>> df.collect(statement_params=statement_params)
     """
     telemetry_type = f"{env.SOURCE.lower()}_{TelemetryField.TYPE_FUNCTION_USAGE.value}"
-    statement_params: Dict[str, Any] = {
+    statement_params: dict[str, Any] = {
         connector_telemetry.TelemetryField.KEY_SOURCE.value: env.SOURCE,
         TelemetryField.KEY_PROJECT.value: project,
         TelemetryField.KEY_SUBPROJECT.value: subproject,
         TelemetryField.KEY_OS.value: env.OS,
-        TelemetryField.KEY_VERSION.value: env.VERSION,
+        TelemetryField.KEY_VERSION.value: snowml_version.VERSION,
         TelemetryField.KEY_PYTHON_VERSION.value: env.PYTHON_VERSION,
         connector_telemetry.TelemetryField.KEY_TYPE.value: telemetry_type,
         TelemetryField.KEY_CATEGORY.value: function_category,
@@ -334,7 +359,7 @@ def get_function_usage_statement_params(
     if api_calls:
         statement_params[TelemetryField.KEY_API_CALLS.value] = []
         for api_call in api_calls:
-            if isinstance(api_call, Dict):
+            if isinstance(api_call, dict):
                 telemetry_api_call = api_call.copy()
                 # convert Callable to str
                 for field, api in api_call.items():
@@ -348,6 +373,10 @@ def get_function_usage_statement_params(
                 statement_params[TelemetryField.KEY_API_CALLS.value].append({TelemetryField.NAME.value: api_call})
     if custom_tags:
         statement_params[TelemetryField.KEY_CUSTOM_TAGS.value] = custom_tags
+    # Snowpark doesn't support None value in statement_params from version 1.29
+    for k in statement_params:
+        if statement_params[k] is None:
+            statement_params[k] = ""
     return statement_params
 
 
@@ -379,16 +408,17 @@ def send_custom_usage(
     *,
     telemetry_type: str,
     subproject: Optional[str] = None,
-    data: Optional[Dict[str, Any]] = None,
+    data: Optional[dict[str, Any]] = None,
     **kwargs: Any,
 ) -> None:
-    active_session = next(iter(session._get_active_sessions()))
-    assert active_session, "Missing active session object"
+    conn = _get_snowflake_connection()
 
-    client = _SourceTelemetryClient(conn=active_session._conn._conn, project=project, subproject=subproject)
-    common_metrics = client._create_basic_telemetry_data(telemetry_type=telemetry_type)
-    data = {**common_metrics, TelemetryField.KEY_DATA.value: data, **kwargs}
-    client._send(msg=data)
+    # Send telemetry if Snowflake connection is available.
+    if conn is not None:
+        client = _SourceTelemetryClient(conn=conn, project=project, subproject=subproject)
+        common_metrics = client._create_basic_telemetry_data(telemetry_type=telemetry_type)
+        data = {**common_metrics, TelemetryField.KEY_DATA.value: data, **kwargs}
+        client._send(msg=data)
 
 
 def send_api_usage_telemetry(
@@ -400,17 +430,17 @@ def send_api_usage_telemetry(
     api_calls_extractor: Optional[
         Callable[
             ...,
-            List[
+            list[
                 Union[
-                    Dict[str, Union[Callable[..., Any], str]],
+                    dict[str, Union[Callable[..., Any], str]],
                     Union[Callable[..., Any], str],
                 ]
             ],
         ]
     ] = None,
-    sfqids_extractor: Optional[Callable[..., List[str]]] = None,
+    sfqids_extractor: Optional[Callable[..., list[str]]] = None,
     subproject_extractor: Optional[Callable[[Any], str]] = None,
-    custom_tags: Optional[Dict[str, Union[bool, int, str, float]]] = None,
+    custom_tags: Optional[dict[str, Union[bool, int, str, float]]] = None,
 ) -> Callable[[Callable[_Args, _ReturnValue]], Callable[_Args, _ReturnValue]]:
     """
     Decorator that sends API usage telemetry and adds function usage statement parameters to the dataframe returned by
@@ -435,16 +465,17 @@ def send_api_usage_telemetry(
 
     # noqa: DAR402
     """
-
     if subproject is not None and subproject_extractor is not None:
         raise ValueError("Specifying both subproject and subproject_extractor is not allowed")
 
     def decorator(func: Callable[_Args, _ReturnValue]) -> Callable[_Args, _ReturnValue]:
         @functools.wraps(func)
         def wrap(*args: Any, **kwargs: Any) -> _ReturnValue:
+            start_time = time.perf_counter()
+
             params = _get_func_params(func, func_params_to_log, args, kwargs) if func_params_to_log else None
 
-            api_calls: List[Union[Dict[str, Union[Callable[..., Any], str]], Callable[..., Any], str]] = []
+            api_calls: list[Union[dict[str, Union[Callable[..., Any], str]], Callable[..., Any], str]] = []
             if api_calls_extractor:
                 extracted_api_calls = api_calls_extractor(args[0])
                 for api_call in extracted_api_calls:
@@ -474,7 +505,7 @@ def send_api_usage_telemetry(
                 custom_tags=custom_tags,
             )
 
-            def update_stmt_params_if_snowpark_df(obj: _ReturnValue, statement_params: Dict[str, Any]) -> _ReturnValue:
+            def update_stmt_params_if_snowpark_df(obj: _ReturnValue, statement_params: dict[str, Any]) -> _ReturnValue:
                 """
                 Update SnowML function usage statement parameters to the object if it is a Snowpark DataFrame.
                 Used to track APIs returning a Snowpark DataFrame.
@@ -504,7 +535,6 @@ def send_api_usage_telemetry(
                 return update_stmt_params_if_snowpark_df(result, statement_params)
 
             # prioritize `conn_attr_name` over the active session
-            telemetry_enabled = True
             if conn_attr_name:
                 # raise AttributeError if conn attribute does not exist in `self`
                 conn = operator.attrgetter(conn_attr_name)(args[0])
@@ -512,16 +542,10 @@ def send_api_usage_telemetry(
                     raise TypeError(
                         f"Expected a conn object of type {' or '.join(_CONNECTION_TYPES.keys())} but got {type(conn)}"
                     )
-            # get an active session
             else:
-                try:
-                    active_session = next(iter(session._get_active_sessions()))
-                    conn = active_session._conn._conn
-                    telemetry_enabled = active_session.telemetry_enabled
-                except snowpark_exceptions.SnowparkSessionException:
-                    conn = None
+                conn = _get_snowflake_connection()
 
-            if conn is None or not telemetry_enabled:
+            if conn is None:
                 # Telemetry not enabled, just execute without our additional telemetry logic
                 try:
                     return ctx.run(execute_func_with_statement_params)
@@ -555,8 +579,16 @@ def send_api_usage_telemetry(
                         )
                 else:
                     me = e
+
                 telemetry_args["error"] = repr(me)
                 telemetry_args["error_code"] = me.error_code
+                # exclude telemetry frames
+                excluded_frames = 2
+                tb = traceback.extract_tb(sys.exc_info()[2])
+                formatted_tb = "".join(traceback.format_list(tb[excluded_frames:]))
+                formatted_exception = traceback.format_exception_only(*sys.exc_info()[:2])[0]  # error type + message
+                telemetry_args["stack_trace"] = formatted_tb + formatted_exception
+
                 me.original_exception._snowflake_ml_handled = True  # type: ignore[attr-defined]
                 if e is not me:
                     raise  # Directly raise non-wrapped exceptions to preserve original stacktrace
@@ -565,6 +597,7 @@ def send_api_usage_telemetry(
                 else:
                     raise me.original_exception from e
             finally:
+                telemetry_args["duration"] = time.perf_counter() - start_time  # type: ignore[assignment]
                 telemetry.send_function_usage_telemetry(**telemetry_args)
                 global _log_counter
                 _log_counter += 1
@@ -595,7 +628,7 @@ def _get_full_func_name(func: Callable[..., Any]) -> str:
 
 def _get_func_params(
     func: Callable[..., Any], func_params_to_log: Optional[Iterable[str]], args: Any, kwargs: Any
-) -> Dict[str, Any]:
+) -> dict[str, Any]:
     """
     Get function parameters.
 
@@ -620,7 +653,7 @@ def _get_func_params(
     return params
 
 
-def _extract_arg_value(field: str, func_spec: inspect.FullArgSpec, args: Any, kwargs: Any) -> Tuple[bool, Any]:
+def _extract_arg_value(field: str, func_spec: inspect.FullArgSpec, args: Any, kwargs: Any) -> tuple[bool, Any]:
     """
     Function to extract a specified argument value.
 
@@ -648,8 +681,10 @@ def _extract_arg_value(field: str, func_spec: inspect.FullArgSpec, args: Any, kw
     elif func_spec.kwonlydefaults and field in func_spec.kwonlyargs:
         if field in kwargs:
             return True, kwargs[field]
-        else:
+        elif field in func_spec.kwonlydefaults:
             return True, func_spec.kwonlydefaults[field]
+        else:
+            return False, None
     else:
         return False, None
 
@@ -683,11 +718,11 @@ class _SourceTelemetryClient:
         self.source: str = env.SOURCE
         self.project: Optional[str] = project
         self.subproject: Optional[str] = subproject
-        self.version = env.VERSION
+        self.version = snowml_version.VERSION
         self.python_version: str = env.PYTHON_VERSION
         self.os: str = env.OS
 
-    def _send(self, msg: Dict[str, Any], timestamp: Optional[int] = None) -> None:
+    def _send(self, msg: dict[str, Any], timestamp: Optional[int] = None) -> None:
         """
         Add telemetry data to a batch in connector client.
 
@@ -701,7 +736,7 @@ class _SourceTelemetryClient:
             telemetry_data = connector_telemetry.TelemetryData(message=msg, timestamp=timestamp)
             self._telemetry.try_add_log_to_batch(telemetry_data)
 
-    def _create_basic_telemetry_data(self, telemetry_type: str) -> Dict[str, Any]:
+    def _create_basic_telemetry_data(self, telemetry_type: str) -> dict[str, Any]:
         message = {
             connector_telemetry.TelemetryField.KEY_SOURCE.value: self.source,
             TelemetryField.KEY_PROJECT.value: self.project,
@@ -718,12 +753,14 @@ class _SourceTelemetryClient:
         self,
         func_name: str,
         function_category: str,
-        func_params: Optional[Dict[str, Any]] = None,
-        api_calls: Optional[List[Dict[str, Any]]] = None,
-        sfqids: Optional[List[Any]] = None,
-        custom_tags: Optional[Dict[str, Union[bool, int, str, float]]] = None,
+        duration: float,
+        func_params: Optional[dict[str, Any]] = None,
+        api_calls: Optional[list[dict[str, Any]]] = None,
+        sfqids: Optional[list[Any]] = None,
+        custom_tags: Optional[dict[str, Union[bool, int, str, float]]] = None,
         error: Optional[str] = None,
         error_code: Optional[str] = None,
+        stack_trace: Optional[str] = None,
     ) -> None:
         """
         Send function usage telemetry message.
@@ -731,14 +768,16 @@ class _SourceTelemetryClient:
         Args:
             func_name: Function name.
             function_category: Function category.
+            duration: Function duration.
             func_params: Function parameters.
             api_calls: API calls.
             sfqids: Snowflake query IDs.
             custom_tags: Custom tags.
             error: Error.
             error_code: Error code.
+            stack_trace: Error stack trace.
         """
-        data: Dict[str, Any] = {
+        data: dict[str, Any] = {
             TelemetryField.KEY_FUNC_NAME.value: func_name,
             TelemetryField.KEY_CATEGORY.value: function_category,
         }
@@ -752,14 +791,16 @@ class _SourceTelemetryClient:
             data[TelemetryField.KEY_CUSTOM_TAGS.value] = custom_tags
 
         telemetry_type = f"{self.source.lower()}_{TelemetryField.TYPE_FUNCTION_USAGE.value}"
-        message: Dict[str, Any] = {
+        message: dict[str, Any] = {
             **self._create_basic_telemetry_data(telemetry_type),
             TelemetryField.KEY_DATA.value: data,
+            TelemetryField.KEY_DURATION.value: duration,
         }
 
         if error:
             message[TelemetryField.KEY_ERROR_INFO.value] = error
             message[TelemetryField.KEY_ERROR_CODE.value] = error_code
+            message[TelemetryField.KEY_STACK_TRACE.value] = stack_trace
 
         self._send(message)
 
@@ -770,7 +811,7 @@ class _SourceTelemetryClient:
             self._telemetry.send_batch()
 
 
-def get_sproc_statement_params_kwargs(sproc: Callable[..., Any], statement_params: Dict[str, Any]) -> Dict[str, Any]:
+def get_sproc_statement_params_kwargs(sproc: Callable[..., Any], statement_params: dict[str, Any]) -> dict[str, Any]:
     """
     Get statement_params keyword argument for sproc call.
 
