@@ -110,11 +110,14 @@ def create_model_metadata(
         python_version=python_version,
         embed_local_ml_library=embed_local_ml_library,
         prefer_pip=prefer_pip,
+        target_platforms=target_platforms,
     )
 
     if embed_local_ml_library:
         env.snowpark_ml_version = f"{snowml_version.VERSION}+{file_utils.hash_directory(path_to_copy)}"
 
+    # Persist full method_options
+    method_options: dict[str, dict[str, Any]] = kwargs.pop("method_options", {})
     model_meta = ModelMetadata(
         name=name,
         env=env,
@@ -123,6 +126,7 @@ def create_model_metadata(
         signatures=signatures,
         function_properties=function_properties,
         task=task,
+        method_options=method_options,
     )
 
     code_dir_path = os.path.join(model_dir_path, MODEL_CODE_DIR)
@@ -162,8 +166,9 @@ def _create_env_for_model_metadata(
     python_version: Optional[str] = None,
     embed_local_ml_library: bool = False,
     prefer_pip: bool = False,
+    target_platforms: Optional[list[model_types.TargetPlatform]] = None,
 ) -> model_env.ModelEnv:
-    env = model_env.ModelEnv(prefer_pip=prefer_pip)
+    env = model_env.ModelEnv(prefer_pip=prefer_pip, target_platforms=target_platforms)
 
     # Mypy doesn't like getter and setter have different types. See python/mypy #3004
     env.conda_dependencies = conda_dependencies  # type: ignore[assignment]
@@ -254,6 +259,7 @@ class ModelMetadata:
         original_metadata_version: Optional[str] = model_meta_schema.MODEL_METADATA_VERSION,
         task: model_types.Task = model_types.Task.UNKNOWN,
         explain_algorithm: Optional[model_meta_schema.ModelExplainAlgorithm] = None,
+        method_options: Optional[dict[str, dict[str, Any]]] = None,
     ) -> None:
         self.name = name
         self.signatures: dict[str, model_signature.ModelSignature] = dict()
@@ -281,6 +287,7 @@ class ModelMetadata:
 
         self.task: model_types.Task = task
         self.explain_algorithm: Optional[model_meta_schema.ModelExplainAlgorithm] = explain_algorithm
+        self.method_options: dict[str, dict[str, Any]] = method_options or {}
 
     @property
     def min_snowpark_ml_version(self) -> str:
@@ -340,6 +347,7 @@ class ModelMetadata:
                     else None
                 ),
                 "function_properties": self.function_properties,
+                "method_options": self.method_options,
             }
         )
         with open(model_yaml_path, "w", encoding="utf-8") as out:
@@ -379,6 +387,7 @@ class ModelMetadata:
             task=loaded_meta.get("task", model_types.Task.UNKNOWN.value),
             explainability=loaded_meta.get("explainability", None),
             function_properties=loaded_meta.get("function_properties", {}),
+            method_options=loaded_meta.get("method_options", {}),
         )
 
     @classmethod
@@ -434,4 +443,5 @@ class ModelMetadata:
             task=model_types.Task(model_dict.get("task", model_types.Task.UNKNOWN.value)),
             explain_algorithm=explanation_algorithm,
             function_properties=model_dict.get("function_properties", {}),
+            method_options=model_dict.get("method_options", {}),
         )
