@@ -1,4 +1,3 @@
-import warnings
 from types import ModuleType
 from typing import Any, Optional, Union, overload
 
@@ -6,15 +5,18 @@ import pandas as pd
 
 from snowflake import snowpark
 from snowflake.ml._internal import telemetry
-from snowflake.ml._internal.utils import sql_identifier
+from snowflake.ml._internal.utils import query_result_checker, sql_identifier
 from snowflake.ml.model import (
     Model,
     ModelVersion,
+    event_handler,
     model_signature,
-    type_hints as model_types,
+    target_platform,
+    task,
+    type_hints,
 )
 from snowflake.ml.model._client.model import model_version_impl
-from snowflake.ml.monitoring import model_monitor, model_monitor_version
+from snowflake.ml.monitoring import model_monitor
 from snowflake.ml.monitoring._manager import model_monitor_manager
 from snowflake.ml.monitoring.entities import model_monitor_config
 from snowflake.ml.registry._manager import model_manager
@@ -30,6 +32,7 @@ _MODEL_MONITORING_DISABLED_ERROR = (
 
 
 class Registry:
+    @telemetry.send_api_usage_telemetry(project=_TELEMETRY_PROJECT, subproject=_MODEL_TELEMETRY_SUBPROJECT)
     def __init__(
         self,
         session: session.Session,
@@ -74,6 +77,32 @@ class Registry:
                 else sql_identifier.SqlIdentifier("PUBLIC")
             )
 
+        database_results = (
+            query_result_checker.SqlResultValidator(
+                session, f"""SHOW DATABASES LIKE '{self._database_name.resolved()}';"""
+            )
+            .has_column("name", allow_empty=True)
+            .validate()
+        )
+
+        db_names = [row["name"] for row in database_results]
+        if not self._database_name.resolved() in db_names:
+            raise ValueError(f"Database {self._database_name} does not exist.")
+
+        schema_results = (
+            query_result_checker.SqlResultValidator(
+                session,
+                f"""SHOW SCHEMAS LIKE '{self._schema_name.resolved()}'
+                IN DATABASE {self._database_name.identifier()};""",
+            )
+            .has_column("name", allow_empty=True)
+            .validate()
+        )
+
+        schema_names = [row["name"] for row in schema_results]
+        if not self._schema_name.resolved() in schema_names:
+            raise ValueError(f"Schema {self._schema_name} does not exist.")
+
         self._model_manager = model_manager.ModelManager(
             session,
             database_name=self._database_name,
@@ -102,7 +131,7 @@ class Registry:
     @overload
     def log_model(
         self,
-        model: model_types.SupportedModelType,
+        model: type_hints.SupportedModelType,
         *,
         model_name: str,
         version_name: Optional[str] = None,
@@ -112,15 +141,15 @@ class Registry:
         pip_requirements: Optional[list[str]] = None,
         artifact_repository_map: Optional[dict[str, str]] = None,
         resource_constraint: Optional[dict[str, str]] = None,
-        target_platforms: Optional[list[model_types.SupportedTargetPlatformType]] = None,
+        target_platforms: Optional[list[Union[target_platform.TargetPlatform, str]]] = None,
         python_version: Optional[str] = None,
         signatures: Optional[dict[str, model_signature.ModelSignature]] = None,
-        sample_input_data: Optional[model_types.SupportedDataType] = None,
+        sample_input_data: Optional[type_hints.SupportedDataType] = None,
         user_files: Optional[dict[str, list[str]]] = None,
         code_paths: Optional[list[str]] = None,
         ext_modules: Optional[list[ModuleType]] = None,
-        task: model_types.Task = model_types.Task.UNKNOWN,
-        options: Optional[model_types.ModelSaveOption] = None,
+        task: task.Task = task.Task.UNKNOWN,
+        options: Optional[type_hints.ModelSaveOption] = None,
     ) -> ModelVersion:
         """
         Log a model with various parameters and metadata, or a ModelVersion object.
@@ -142,20 +171,27 @@ class Registry:
                 to specify a dependency. It is a recommended way to specify your dependencies using conda. When channel
                 is not specified, Snowflake Anaconda Channel will be used. Defaults to None.
             pip_requirements: List of Pip package specifications. Defaults to None.
-                Models with pip requirements are currently only runnable in Snowpark Container Services.
-                See https://docs.snowflake.com/en/developer-guide/snowflake-ml/model-registry/container for more.
-                Models with pip requirements specified will not be executable in Snowflake Warehouse where all
-                dependencies must be retrieved from Snowflake Anaconda Channel.
+                Models running in a Snowflake Warehouse must also specify a pip artifact repository (see
+                `artifact_repository_map`). Otherwise, models with pip requirements are runnable only in Snowpark
+                Container Services. See
+                https://docs.snowflake.com/en/developer-guide/snowflake-ml/model-registry/container for more.
             artifact_repository_map: Specifies a mapping of package channels or platforms to custom artifact
-                repositories. Defaults to None. Currently, the mapping applies only to warehouse execution.
-                Note : This feature is currently in Private Preview; please contact your Snowflake account team
-                to enable it.
+                repositories. Defaults to None. Currently, the mapping applies only to Warehouse execution.
+                Note : This feature is currently in Public Preview.
                 Format: {channel_name: artifact_repository_name}, where:
-                   - channel_name: The name of the Conda package channel (e.g., 'condaforge') or 'pip' for pip packages.
-                   - artifact_repository_name: The name or URL of the repository to fetch packages from.
+                   - channel_name: Currently must be 'pip'.
+                   - artifact_repository_name: The identifier of the artifact repository to fetch packages from, e.g.
+                     `snowflake.snowpark.pypi_shared_repository`.
             resource_constraint: Mapping of resource constraint keys and values, e.g. {"architecture": "x86"}.
             target_platforms: List of target platforms to run the model. The only acceptable inputs are a combination of
-                {"WAREHOUSE", "SNOWPARK_CONTAINER_SERVICES"}. Defaults to None.
+                "WAREHOUSE" and "SNOWPARK_CONTAINER_SERVICES", or a target platform constant:
+                - ["WAREHOUSE"] or snowflake.ml.model.target_platform.WAREHOUSE_ONLY (Warehouse only)
+                - ["SNOWPARK_CONTAINER_SERVICES"] or
+                  snowflake.ml.model.target_platform.SNOWPARK_CONTAINER_SERVICES_ONLY
+                  (Snowpark Container Services only)
+                - ["WAREHOUSE", "SNOWPARK_CONTAINER_SERVICES"] or
+                  snowflake.ml.model.target_platform.BOTH_WAREHOUSE_AND_SNOWPARK_CONTAINER_SERVICES (Both)
+                Defaults to None. When None, the target platforms will be both.
             python_version: Python version in which the model is run. Defaults to None.
             signatures: Model data signatures for inputs and outputs for various target methods. If it is None,
                 sample_input_data would be used to infer the signatures for those models that cannot automatically
@@ -185,7 +221,8 @@ class Registry:
                 - target_methods: List of target methods to register when logging the model.
                   This option is not used in MLFlow models. Defaults to None, in which case the model handler's
                   default target methods will be used.
-                - save_location: Location to save the model and metadata.
+                - save_location: Local directory to save the the serialized model files first before
+                  uploading to Snowflake. This is useful when default tmp directory is not writable.
                 - method_options: Per-method saving options. This dictionary has method names as keys and dictionary
                     values with the desired options.
 
@@ -198,6 +235,10 @@ class Registry:
                     - max_batch_size: Maximum batch size that the method could accept in the Snowflake Warehouse.
                         Defaults to None, determined automatically by Snowflake.
                     - function_type: One of supported model method function types (FUNCTION or TABLE_FUNCTION).
+                    - volatility: Volatility level for the function (use Volatility.VOLATILE or Volatility.IMMUTABLE).
+                        Volatility.VOLATILE functions may return different results for the same arguments, while
+                        Volatility.IMMUTABLE functions always return the same result for the same arguments.
+                        Defaults to None (no volatility specified).
         Returns:
             ModelVersion: ModelVersion object corresponding to the model just logged.
         """
@@ -242,7 +283,7 @@ class Registry:
     )
     def log_model(
         self,
-        model: Union[model_types.SupportedModelType, ModelVersion],
+        model: Union[type_hints.SupportedModelType, ModelVersion],
         *,
         model_name: str,
         version_name: Optional[str] = None,
@@ -252,15 +293,15 @@ class Registry:
         pip_requirements: Optional[list[str]] = None,
         artifact_repository_map: Optional[dict[str, str]] = None,
         resource_constraint: Optional[dict[str, str]] = None,
-        target_platforms: Optional[list[model_types.SupportedTargetPlatformType]] = None,
+        target_platforms: Optional[list[Union[target_platform.TargetPlatform, str]]] = None,
         python_version: Optional[str] = None,
         signatures: Optional[dict[str, model_signature.ModelSignature]] = None,
-        sample_input_data: Optional[model_types.SupportedDataType] = None,
+        sample_input_data: Optional[type_hints.SupportedDataType] = None,
         user_files: Optional[dict[str, list[str]]] = None,
         code_paths: Optional[list[str]] = None,
         ext_modules: Optional[list[ModuleType]] = None,
-        task: model_types.Task = model_types.Task.UNKNOWN,
-        options: Optional[model_types.ModelSaveOption] = None,
+        task: task.Task = task.Task.UNKNOWN,
+        options: Optional[type_hints.ModelSaveOption] = None,
     ) -> ModelVersion:
         """
         Log a model with various parameters and metadata, or a ModelVersion object.
@@ -282,20 +323,27 @@ class Registry:
                 to specify a dependency. It is a recommended way to specify your dependencies using conda. When channel
                 is not specified, Snowflake Anaconda Channel will be used. Defaults to None.
             pip_requirements: List of Pip package specifications. Defaults to None.
-                Models with pip requirements are currently only runnable in Snowpark Container Services.
-                See https://docs.snowflake.com/en/developer-guide/snowflake-ml/model-registry/container for more.
-                Models with pip requirements specified will not be executable in Snowflake Warehouse where all
-                dependencies must be retrieved from Snowflake Anaconda Channel.
+                Models running in a Snowflake Warehouse must also specify a pip artifact repository (see
+                `artifact_repository_map`). Otherwise, models with pip requirements are runnable only in Snowpark
+                Container Services. See
+                https://docs.snowflake.com/en/developer-guide/snowflake-ml/model-registry/container for more.
             artifact_repository_map: Specifies a mapping of package channels or platforms to custom artifact
-                repositories. Defaults to None. Currently, the mapping applies only to warehouse execution.
-                Note : This feature is currently in Private Preview; please contact your Snowflake account team to
-                enable it.
+                repositories. Defaults to None. Currently, the mapping applies only to Warehouse execution.
+                Note : This feature is currently in Public Preview.
                 Format: {channel_name: artifact_repository_name}, where:
-                   - channel_name: The name of the Conda package channel (e.g., 'condaforge') or 'pip' for pip packages.
-                   - artifact_repository_name: The name or URL of the repository to fetch packages from.
+                   - channel_name: Currently must be 'pip'.
+                   - artifact_repository_name: The identifier of the artifact repository to fetch packages from, e.g.
+                     `snowflake.snowpark.pypi_shared_repository`.
             resource_constraint: Mapping of resource constraint keys and values, e.g. {"architecture": "x86"}.
             target_platforms: List of target platforms to run the model. The only acceptable inputs are a combination of
-                {"WAREHOUSE", "SNOWPARK_CONTAINER_SERVICES"}. Defaults to None.
+                "WAREHOUSE" and "SNOWPARK_CONTAINER_SERVICES", or a target platform constant:
+                - ["WAREHOUSE"] or snowflake.ml.model.target_platform.WAREHOUSE_ONLY (Warehouse only)
+                - ["SNOWPARK_CONTAINER_SERVICES"] or
+                  snowflake.ml.model.target_platform.SNOWPARK_CONTAINER_SERVICES_ONLY
+                  (Snowpark Container Services only)
+                - ["WAREHOUSE", "SNOWPARK_CONTAINER_SERVICES"] or
+                  snowflake.ml.model.target_platform.BOTH_WAREHOUSE_AND_SNOWPARK_CONTAINER_SERVICES (Both)
+                Defaults to None. When None, the target platforms will be both.
             python_version: Python version in which the model is run. Defaults to None.
             signatures: Model data signatures for inputs and outputs for various target methods. If it is None,
                 sample_input_data would be used to infer the signatures for those models that cannot automatically
@@ -322,6 +370,12 @@ class Registry:
                     Warehouse. It detects any ==x.y.z in specifiers and replaced with >=x.y, <(x+1). Defaults to True.
                 - function_type: Set the method function type globally. To set method function types individually see
                   function_type in model_options.
+                - volatility: Set the volatility for all model methods globally (use Volatility.VOLATILE or
+                  Volatility.IMMUTABLE). Volatility.VOLATILE functions may return different results for the same
+                  arguments, while Volatility.IMMUTABLE functions always return the same result for the same arguments.
+                  Defaults are set automatically based on model type: supported models (sklearn, xgboost, pytorch,
+                  huggingface_pipeline, mlflow, etc.) default to IMMUTABLE, while custom models default to VOLATILE.
+                  Individual method volatility can be set in method_options and will override this global setting.
                 - target_methods: List of target methods to register when logging the model.
                   This option is not used in MLFlow models. Defaults to None, in which case the model handler's
                   default target methods will be used.
@@ -338,9 +392,15 @@ class Registry:
                     - max_batch_size: Maximum batch size that the method could accept in the Snowflake Warehouse.
                         Defaults to None, determined automatically by Snowflake.
                     - function_type: One of supported model method function types (FUNCTION or TABLE_FUNCTION).
+                    - volatility: Volatility level for the function (use Volatility.VOLATILE or Volatility.IMMUTABLE).
+                        Volatility.VOLATILE functions may return different results for the same arguments, while
+                        Volatility.IMMUTABLE functions always return the same result for the same arguments.
+                        This per-method setting overrides any global volatility setting.
+                        Defaults to None (no volatility specified).
 
         Raises:
             ValueError: If extra arguments are specified ModelVersion is provided.
+            Exception: If the model logging fails.
 
         Returns:
             ModelVersion: ModelVersion object corresponding to the model just logged.
@@ -393,38 +453,44 @@ class Registry:
                     raise ValueError(
                         "When calling log_model with a ModelVersion, only model_name and version_name may be specified."
                     )
-            if task is not model_types.Task.UNKNOWN:
+            if task is not type_hints.Task.UNKNOWN:
                 raise ValueError("`task` cannot be specified when calling log_model with a ModelVersion.")
 
-        if pip_requirements:
-            warnings.warn(
-                "Models logged specifying `pip_requirements` can not be executed "
-                "in Snowflake Warehouse where all dependencies are required to be retrieved "
-                "from Snowflake Anaconda Channel.",
-                category=UserWarning,
-                stacklevel=1,
-            )
-        return self._model_manager.log_model(
-            model=model,
-            model_name=model_name,
-            version_name=version_name,
-            comment=comment,
-            metrics=metrics,
-            conda_dependencies=conda_dependencies,
-            pip_requirements=pip_requirements,
-            artifact_repository_map=artifact_repository_map,
-            resource_constraint=resource_constraint,
-            target_platforms=target_platforms,
-            python_version=python_version,
-            signatures=signatures,
-            sample_input_data=sample_input_data,
-            user_files=user_files,
-            code_paths=code_paths,
-            ext_modules=ext_modules,
-            task=task,
-            options=options,
-            statement_params=statement_params,
-        )
+        registry_event_handler = event_handler.ModelEventHandler()
+        with registry_event_handler.status("Logging model", total=6) as status:
+            # Step 1: Validation and setup
+            status.update("validating model and dependencies...")
+            status.increment()
+
+            # Perform the actual model logging
+            try:
+                result = self._model_manager.log_model(
+                    model=model,
+                    model_name=model_name,
+                    version_name=version_name,
+                    comment=comment,
+                    metrics=metrics,
+                    conda_dependencies=conda_dependencies,
+                    pip_requirements=pip_requirements,
+                    artifact_repository_map=artifact_repository_map,
+                    resource_constraint=resource_constraint,
+                    target_platforms=target_platforms,
+                    python_version=python_version,
+                    signatures=signatures,
+                    sample_input_data=sample_input_data,
+                    user_files=user_files,
+                    code_paths=code_paths,
+                    ext_modules=ext_modules,
+                    task=task,
+                    options=options,
+                    statement_params=statement_params,
+                    progress_status=status,
+                )
+                status.update(label="Model logged successfully.", state="complete", expanded=False)
+                return result
+            except Exception as e:
+                status.update(label="Model logging failed.", state="error", expanded=False)
+                raise e
 
     @telemetry.send_api_usage_telemetry(
         project=_TELEMETRY_PROJECT,
@@ -499,7 +565,6 @@ class Registry:
         project=telemetry.TelemetryProject.MLOPS.value,
         subproject=telemetry.TelemetrySubProject.MONITORING.value,
     )
-    @snowpark._internal.utils.private_preview(version=model_monitor_version.SNOWFLAKE_ML_MONITORING_MIN_VERSION)
     def add_monitor(
         self,
         name: str,
@@ -524,7 +589,7 @@ class Registry:
         return self._model_monitor_manager.add_monitor(name, source_config, model_monitor_config)
 
     @overload
-    def get_monitor(self, model_version: model_version_impl.ModelVersion) -> model_monitor.ModelMonitor:
+    def get_monitor(self, *, model_version: model_version_impl.ModelVersion) -> model_monitor.ModelMonitor:
         """Get a Model Monitor on a Model Version from the Registry.
 
         Args:
@@ -533,7 +598,7 @@ class Registry:
         ...
 
     @overload
-    def get_monitor(self, name: str) -> model_monitor.ModelMonitor:
+    def get_monitor(self, *, name: str) -> model_monitor.ModelMonitor:
         """Get a Model Monitor by name from the Registry.
 
         Args:
@@ -545,7 +610,6 @@ class Registry:
         project=telemetry.TelemetryProject.MLOPS.value,
         subproject=telemetry.TelemetrySubProject.MONITORING.value,
     )
-    @snowpark._internal.utils.private_preview(version=model_monitor_version.SNOWFLAKE_ML_MONITORING_MIN_VERSION)
     def get_monitor(
         self, *, name: Optional[str] = None, model_version: Optional[model_version_impl.ModelVersion] = None
     ) -> model_monitor.ModelMonitor:
@@ -574,7 +638,6 @@ class Registry:
         project=telemetry.TelemetryProject.MLOPS.value,
         subproject=telemetry.TelemetrySubProject.MONITORING.value,
     )
-    @snowpark._internal.utils.private_preview(version=model_monitor_version.SNOWFLAKE_ML_MONITORING_MIN_VERSION)
     def show_model_monitors(self) -> list[snowpark.Row]:
         """Show all model monitors in the registry.
 
@@ -592,7 +655,6 @@ class Registry:
         project=telemetry.TelemetryProject.MLOPS.value,
         subproject=telemetry.TelemetrySubProject.MONITORING.value,
     )
-    @snowpark._internal.utils.private_preview(version=model_monitor_version.SNOWFLAKE_ML_MONITORING_MIN_VERSION)
     def delete_monitor(self, name: str) -> None:
         """Delete a Model Monitor by name from the Registry.
 
